@@ -1,27 +1,42 @@
 import sys,os
 
-if sys.version_info < (3, 8):
-    print("only support python >= 3.8, but now is {}".format(sys.version_info))
+if sys.version_info < (3, 7):
+    print("only support python >= 3.7, but now is {}".format(sys.version_info))
     sys.exit(1)
 
+# Init lanuage first to ensure use function _  works correctly
+try:
+    import parameters
+    import i18n
+except ImportError:
+    from COMTool import parameters
+    from COMTool import i18n
+
+def loadConfig():
+    paramObj = parameters.Parameters()
+    paramObj.load(parameters.configFilePath)
+    return paramObj
+
+print("-- loading config from", parameters.configFilePath)
+programConfig = loadConfig()
+print("-- loading config complete")
+i18n.set_locale(programConfig.basic["locale"])
 
 try:
-    import parameters,helpAbout,autoUpdate
+    import helpAbout,autoUpdate
     from Combobox import ComboBox
-    import i18n
     from i18n import _
     import version
     import utils_ui
-    from conn.conn_serial import Serial
+    from conn import Serial, ConnectionStatus, TCP_UDP
     from plugins import plugins
     from widgets import TitleBar, CustomTitleBarWindowMixin, EventFilter
 except ImportError:
-    from COMTool import parameters,helpAbout,autoUpdate, utils_ui
+    from COMTool import helpAbout,autoUpdate, utils_ui
     from COMTool.Combobox import ComboBox
-    from COMTool import i18n
     from COMTool.i18n import _
     from COMTool import version
-    from COMTool.conn.conn_serial import Serial
+    from COMTool.conn import Serial, ConnectionStatus, TCP_UDP
     from COMTool.plugins import plugins
     from .widgets import TitleBar, CustomTitleBarWindowMixin, EventFilter
 
@@ -53,46 +68,57 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
     app = None
     needRestart = False
 
-    def __init__(self,app, eventFilter):
+    def __init__(self,app, eventFilter, config):
         QMainWindow.__init__(self)
         self.app = app
         self.eventFilter = eventFilter
         self.DataPath = parameters.dataPath
-        self.config = self.loadConfig()
-        i18n.set_locale(self.config.basic["locale"])
+        self.config = config
         self.initVar()
-        self.initConn(self.config.basic["connId"], self.config.conns)
         self.initPlugins(self.config.basic["plugins"], self.config.basic["activePlugin"], self.config.plugins)
         self.initWindow()
+        self.initConn(self.config.basic["connId"], self.config.conns)
         self.uiLoadConfigs(self.config)
         self.initEvent()
         self.uiInitDone()
 
-    def __del__(self):
-        pass
+    def initVar(self):
+        self.strings = parameters.Strings(self.config.basic["locale"])
+        self.dataToSend = []
+        self.fileToSend = []
+        self.connSelectComboboxes = []
+        self.changingConn = False
+        self.connectionWidget = None
 
     def initConn(self, connId, configs):
         # get all conn info
-        self.connections = [Serial()]
+        self.connections = [Serial(), TCP_UDP()]
+        # add conn select combobox
+        self.changingConn = True
+        for i, conn in enumerate(self.connections):
+            for combobox in self.connSelectComboboxes:
+                combobox.addItem(conn.name)
+                if conn.id == connId:
+                    combobox.setCurrentIndex(i)
+        self.changingConn = False
         # init connections
-        for conn in self.connections:
+        activeConn = self.connections[0]
+        for i, conn in enumerate(self.connections):
             conn.onReceived = self.onReceived
             conn.configGlobal = self.config.basic
-            conn.hintSignal = self.hintSignal
+            conn.hintSignal.connect(self.showHint)
+            conn.onConnectionStatus.connect(self.onShowConnStatus)
             config = {}
             if conn.id in configs:
                 config = configs[conn.id]
             else: # add new dict obj for conn to use
                 configs[conn.id] = config
             conn.onInit(config)
-        # init last used one
-        self.connection = None
-        for conn in self.connections:
+            conn.onWidget()
             if conn.id == connId:
-                self.connection = conn
-        if not self.connection:
-            self.connection = self.connections[0]
-
+                activeConn = conn
+        # init last used one
+        self.changeConn(activeConn)
 
     def initPlugins(self, enabled, activeId, configs):
         if not enabled:
@@ -102,10 +128,13 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
         for plugin in self.plugins:
             plugin.hintSignal = self.hintSignal
             plugin.send = self.sendData
+            plugin.isConnected = self.isConnected
             plugin.clearCountSignal = self.clearCountSignal
             plugin.reloadWindowSignal = self.reloadWindowSignal
             plugin.configGlobal = self.config.basic
-            config = {}
+            config = {
+                "plugin_id": plugin.id
+            }
             if plugin.id in configs:
                 config = configs[plugin.id]
             else: # add new dict obj for plugin to use
@@ -128,28 +157,22 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
         plugin.active = True
         parent = None
         if (not "main" in plugin.connParent) and plugin.connParent:
-            plugin.isConnected = self.connection.isConnected
             for pluginParent in self.plugins:
                 if pluginParent.id == plugin.connParent:
                     pluginParent.active = True
-                    pluginParent.isConnected = self.connection.isConnected
                     pluginParent.send = self.sendData
-                    pluginParent.connChilds.append(plugin)
+                    if not plugin in pluginParent.connChilds:
+                        pluginParent.connChilds.append(plugin)
                     parent = pluginParent
                     break
             plugin.send = parent.sendData
         if not parent:
-            plugin.isConnected = self.connection.isConnected
             plugin.send = self.sendData
         self.config.basic["activePlugin"] = plugin.id
-    
-    def initVar(self):
-        self.strings = parameters.Strings(self.config.basic["locale"])
-        self.dataToSend = []
-        self.fileToSend = []
 
     def uiInitDone(self):
-        self.connection.onUiInitDone()
+        for conn in self.connections:
+            conn.onUiInitDone()
         for plugin in self.plugins:
             plugin.onUiInitDone()
         # always hide functional, and show before init complete so we can get its width height
@@ -213,10 +236,21 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
             settingLayout = QVBoxLayout()
             settingWidget.setLayout(settingLayout)
             #    connection settings
+            connSelectCommbox = ComboBox()
+            self.connSelectComboboxes.append(connSelectCommbox)
+            connSelectCommbox.currentIndexChanged.connect(self.onConnChanged)
             connSettingsGroupBox = QGroupBox(_("Connection"))
             layout = QVBoxLayout()
             connSettingsGroupBox.setLayout(layout)
-            layout.addWidget(connectionWidget)
+            layout.addWidget(connSelectCommbox)
+            layout.setContentsMargins(1, 6, 0, 0)
+            connectionWidgetWrapper = QWidget()
+            layout2 = QVBoxLayout()
+            layout2.setContentsMargins(0, 0, 0, 0)
+            connectionWidgetWrapper.setLayout(layout2)
+            layout.addWidget(connectionWidgetWrapper)
+            if connectionWidget:
+                layout2.addWidget(connectionWidget)
             settingLayout.addWidget(connSettingsGroupBox)
             #  other settings
             widget = plugin.onWidgetSettings(settingLayout)
@@ -224,14 +258,14 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
             settingLayout.setContentsMargins(0,0,0,0)
 
             # right functional layout
-            functionalWiget = plugin.onWidgetFunctional(contentWidget)
+            functionalWidget = plugin.onWidgetFunctional(contentWidget)
             contentWidget.addWidget(settingWidget)
             contentWidget.addWidget(mainWidget)
-            contentWidget.addWidget(functionalWiget)
+            contentWidget.addWidget(functionalWidget)
             contentWidget.setStretchFactor(0, 1)
             contentWidget.setStretchFactor(1, 2)
             contentWidget.setStretchFactor(2, 1)
-            return contentWidget, connSettingsGroupBox
+            return contentWidget, connectionWidgetWrapper
         self.tabWidget = QTabWidget()
         # self.contentLayout.setSpacing(0)
         # self.contentLayout.setContentsMargins(0,0,0,0)
@@ -259,17 +293,13 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
         self.contentLayout.addWidget(self.tabWidget)
         # get widgets from plugins
             # widget main
-        self.connectionWidget = self.connection.onWidget()
         self.tabWidgets = []
         self.connParentWidgets = []
         for i, plugin in enumerate(self.plugins):
             active = False
             if plugin.id == self.config.basic["activePlugin"]:
                 active = True
-                conn = self.connectionWidget
-            else:
-                conn = None
-            w, connParent = addTabPanel(self.tabWidget, plugin.name, plugin, conn)
+            w, connParent = addTabPanel(self.tabWidget, plugin.name, plugin, None)
             self.tabWidgets.append(w)
             self.connParentWidgets.append(connParent)
             if active:
@@ -304,7 +334,8 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
         print("config file path:",parameters.configFilePath)
 
     def add_new_window(self):
-        mainWindow = MainWindow(self.app, self.eventFilter)
+        import copy
+        mainWindow = MainWindow(self.app, self.eventFilter, copy.deepcopy(self.config))
         self.eventFilter.listenWindow(mainWindow)
         g_all_windows.append(mainWindow)
 
@@ -375,16 +406,49 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
 
     def changeConnToTab(self, idx):
         print("-- switch to tab", idx)
+        # remove from old container
         parent = self.connectionWidget.parentWidget()
-        if parent.layout().indexOf(self.connectionWidget) >= 0:
-            parent.layout().removeWidget(self.connectionWidget)
+        parent.layout().removeWidget(self.connectionWidget)
+        # add to new container
         newParent = self.connParentWidgets[idx]
         newParent.layout().addWidget(self.connectionWidget)
         self.updateStyle(parent)
         self.updateStyle(newParent)
         self.activePlugin(self.plugins[idx])
 
-    def sendData(self, data_bytes=None, file_path=None, callback=lambda ok,msg:None):
+    def changeConn(self, connection:object):
+        # remove conn widget from the container
+        parent = None
+        if self.connectionWidget:
+            parent = self.connectionWidget.parentWidget()
+            parent.layout().removeWidget(self.connectionWidget)
+            self.connectionWidget.setParent(None)
+            self.updateStyle(parent)
+        # add new conn widget to container
+        self.connection = connection
+        self.connectionWidget = connection.widget
+        if not parent:
+            parent = self.connParentWidgets[self.tabWidget.currentIndex()]
+        parent.layout().addWidget(self.connectionWidget)
+        self.updateStyle(parent)
+
+    def onConnChanged(self, idx):
+        '''
+            combobox changed item
+        '''
+        if self.changingConn:
+            return
+        self.changingConn = True
+        # update all tab's connSelectCombobox widget show
+        for w in self.connSelectComboboxes:
+            w.setCurrentIndex(idx)
+        # change connection
+        print("-- change conn to", self.connections[idx].id)
+        self.changeConn(self.connections[idx])
+        self.config.basic["connId"] = self.connections[idx].id
+        self.changingConn = False
+
+    def sendData(self, data_bytes=None, file_path=None, callback=lambda ok, msg:None):
         if data_bytes:
             self.dataToSend.insert(0, (data_bytes, callback))
         if file_path:
@@ -402,6 +466,8 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
             if plugin.active:
                 plugin.onReceived(data)
 
+    def isConnected(self):
+        return self.connection.isConnected()
 
     def sendDataProcess(self):
         self.receiveProgressStop = False
@@ -442,6 +508,14 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
         self.statusBarSendCount.setText('{}({}): {}'.format(_("Sent"), _("bytes"), send))
         self.statusBarReceiveCount.setText('{}({}): {}'.format(_("Received"), _("bytes"), receive))
 
+    def onShowConnStatus(self, status, msg):
+        if status == ConnectionStatus.CONNECTED:
+            self.onstatusBarText("info", msg)
+        elif status == ConnectionStatus.CLOSED:
+            self.onstatusBarText("info", msg)
+        else:
+            self.onstatusBarText("warning", msg)
+
     def onstatusBarText(self, msg_type, msg):
         if msg_type == "info":
             color = "#008200"
@@ -452,7 +526,7 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
         else:
             color = "#008200"
         text = '<font color={}>{}</font>'.format(color, msg)
-        self.statusBarStauts.setText(text)        
+        self.statusBarStauts.setText(text)
 
     def updateStyle(self, widget):
         self.frameWidget.style().unpolish(widget)
@@ -516,13 +590,9 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
             event.ignore()
 
     def saveConfig(self):
+        # print("save config:", self.config)
         self.config.save(parameters.configFilePath)
-
-    def loadConfig(self):
-        paramObj = parameters.Parameters()
-        paramObj.load(parameters.configFilePath)
-        self.config = paramObj
-        return self.config
+        print("save config compelte")
 
     def uiLoadConfigs(self, config):
         config = config.basic
@@ -618,7 +688,7 @@ class MainWindow(CustomTitleBarWindowMixin, QMainWindow):
         elif result == 1:
             self.config.basic["skipVersion"] = versionInt
 
-            
+
 
     def autoUpdateDetect(self):
         auto = autoUpdate.AutoUpdate()
@@ -655,7 +725,7 @@ def main():
                 gen_tranlate_files(curr_dir)
             app = QApplication(sys.argv)
             eventFilter = EventFilter()
-            mainWindow = MainWindow(app, eventFilter)
+            mainWindow = MainWindow(app, eventFilter, programConfig)
             eventFilter.listenWindow(mainWindow)
             app.installEventFilter(eventFilter)
             g_all_windows.append(mainWindow)
@@ -673,6 +743,7 @@ def main():
             t.start()
             ret = app.exec_()
             if not mainWindow.needRestart:
+                app.removeEventFilter(eventFilter)
                 print("-- no need to restart, now exit")
                 break
     except Exception as e:
