@@ -38,6 +38,7 @@ class Serial(COMM):
             getConfig
     '''
     id = "serial"
+    name = _("Serial")
     showSerialComboboxSignal = pyqtSignal(list)
     showSwitchSignal = pyqtSignal(ConnectionStatus)
     def onInit(self, config):
@@ -93,7 +94,7 @@ class Serial(COMM):
         self.detectSerialPort()
 
     def onWidget(self):
-        serialSetting = QWidget()
+        self.widget = QWidget()
         serialSettingsLayout = QGridLayout()
         serialPortLabek = QLabel(_("Port"))
         serailBaudrateLabel = QLabel(_("Baudrate"))
@@ -152,7 +153,7 @@ class Serial(COMM):
         serialSettingsLayout.addWidget(self.checkBoxRTS, 6, 0,1,1)
         serialSettingsLayout.addWidget(self.checkBoxDTR, 6, 1,1,1)
         serialSettingsLayout.addWidget(self.serialOpenCloseButton, 7, 0,1,2)
-        serialSetting.setLayout(serialSettingsLayout)
+        self.widget.setLayout(serialSettingsLayout)
         self.widgetConfMap["port"]       = self.serialPortCombobox
         self.widgetConfMap["baudrate"]    = self.serailBaudrateCombobox
         self.widgetConfMap["bytesize"]    = self.serailBytesCombobox
@@ -162,8 +163,7 @@ class Serial(COMM):
         self.widgetConfMap["rts"]         = self.checkBoxRTS
         self.widgetConfMap["dtr"]         = self.checkBoxDTR
         self.initEvet()
-        self.widget = serialSetting
-        return serialSetting
+        return self.widget
 
     def initEvet(self):
         self.serialPortCombobox.clicked.connect(self.detectSerialPort)
@@ -414,6 +414,7 @@ class Serial(COMM):
                         print("-- reopen serial")
                         waitingReconnect = False
                         self.onConnectionStatus.emit(ConnectionStatus.CONNECTED, _("Reconnected"))
+                        self.showSwitchSignal.emit(ConnectionStatus.CONNECTED)
                         continue
                     except Exception as e:
                         pass
@@ -443,7 +444,8 @@ class Serial(COMM):
                         pass
                     waitingReconnect = True
                     self.onConnectionStatus.emit(ConnectionStatus.LOSE, _("Connection lose!"))
-                    
+                    self.showSwitchSignal.emit(ConnectionStatus.LOSE)
+
 
     def send(self, data : bytes):
         self.com.write(data)
@@ -459,6 +461,7 @@ if __name__ == "__main__":
     app = QApplication(sys.argv)
 
     conn = Serial()
+    conn.onInit({})
 
     def onReceived(data):
         print(data)
@@ -466,6 +469,5 @@ if __name__ == "__main__":
 
     conn.onReceived = onReceived
     window = conn.onWidget()
-    conn.updateConfig({})
     window.show()
     ret = app.exec_()
