@@ -1,9 +1,9 @@
 
-from PyQt5.QtCore import QObject, Qt, pyqtSignal
+from PyQt5.QtCore import QObject, Qt, pyqtSignal, QEvent
 from PyQt5.QtWidgets import (QApplication, QWidget,QPushButton,QMessageBox,QDesktopWidget,QMainWindow,
                              QVBoxLayout,QHBoxLayout,QGridLayout,QLabel,QRadioButton,QCheckBox,
                              QLineEdit,QGroupBox,QSplitter,QFileDialog, QScrollArea, QInputDialog, QDialog)
-from PyQt5.QtGui import QIcon,QFont,QTextCursor,QPixmap,QColor, QFontMetricsF
+from PyQt5.QtGui import QIcon,QFont,QTextCursor,QPixmap,QColor, QFontMetricsF, QKeySequence, QFocusEvent
 import qtawesome as qta # https://github.com/spyder-ide/qtawesome
 
 try:
@@ -29,41 +29,55 @@ except Exception:
     from . import crc
     from .protocols import defaultProtocols
 
-import os, json
+import os, json, time
 from struct import unpack, pack
 
 class EditRemarDialog(QDialog):
-    def __init__(self, remark = "", icon=None) -> None:
+    def __init__(self, remark = "", icon=None, shortcut = []) -> None:
         super().__init__()
         self.remark = remark
         self.icon = icon
         self.ok = False
+        self.settingShortcut = False
+        self.shortcut = shortcut
 
         layout = QGridLayout()
         self.setLayout(layout)
         layout.addWidget(QLabel(_("Input remark")), 0, 0, 1, 1)
-        remarkInput = QLineEdit(self.remark)
+        self.remarkInput = QLineEdit(self.remark)
         self.iconBtn = QPushButton(self.remark)
         if self.icon:
             self.iconBtn.setIcon(qta.icon(self.icon, color="white"))
-        layout.addWidget(remarkInput, 0, 1, 1, 1)
+        if self.shortcut:
+            name = "+".join([str(name) for v,name in self.shortcut])
+        else:
+            name = _("Record")
+        self.shortcutBtn = QPushButton(name)
+        self.shortcutBtn.setFocusPolicy(Qt.NoFocus)
+        layout.addWidget(self.remarkInput, 0, 1, 1, 1)
         layout.addWidget(QLabel(_("Select icon")), 1, 0, 1, 1)
         layout.addWidget(self.iconBtn, 1, 1, 1, 1)
-        okBtn = QPushButton(_("OK"))
-        cancelBtn = QPushButton(_("Cancel"))
-        layout.addWidget(okBtn, 2, 0, 1, 1)
-        layout.addWidget(cancelBtn, 2, 1, 1, 1)
+        layout.addWidget(QLabel(_("Shortcut")), 2, 0, 1, 1)
+        layout.addWidget(self.shortcutBtn, 2, 1, 1, 1)
+        self.shortcutHint = QLabel(_("Press key to record, or click Cancel"))
+        self.shortcutHint.hide()
+        layout.addWidget(self.shortcutHint, 3, 0, 1, 2)
+        self.okBtn = QPushButton(_("OK"))
+        self.cancelBtn = QPushButton(_("Cancel"))
+        layout.addWidget(self.okBtn, 4, 0, 1, 1)
+        layout.addWidget(self.cancelBtn, 4, 1, 1, 1)
 
         def ok():
             self.ok = True
             self.close()
-        okBtn.clicked.connect(lambda : ok())
-        cancelBtn.clicked.connect(lambda : self.close())
+        self.okBtn.clicked.connect(lambda : ok())
+        self.cancelBtn.clicked.connect(lambda : self.close())
         def updateRemark(text):
             self.remark = text
             self.iconBtn.setText(self.remark)
-        remarkInput.textChanged.connect(updateRemark)
+        self.remarkInput.textChanged.connect(updateRemark)
         self.iconBtn.clicked.connect(lambda: self.selectIcon())
+        self.shortcutBtn.clicked.connect(self.setShortcut)
 
     def selectIcon(self):
         self.icon = selectIcon(parent = self, title = _("Select icon"), btnName = _("OK"), color = utils_ui.getStyleVar("iconSelectorColor"))
@@ -74,9 +88,68 @@ class EditRemarDialog(QDialog):
 
     def exec(self):
         super().exec()
-        return self.ok, self.remark, self.icon
+        return self.ok, self.remark, self.icon, self.shortcut
+
+    def setShortcut(self):
+        if not self.settingShortcut:
+            self.onRecordShortcut()
+        else:
+            self.onRecordShortcutEnd()
+
+    def onRecordShortcut(self):
+        self.shortcut = []
+        self.remarkInput.setEnabled(False)
+        self.iconBtn.setEnabled(False)
+        self.okBtn.setEnabled(False)
+        self.cancelBtn.setEnabled(False)
+        self.shortcutBtn.setText(_("Cancel"))
+        self.shortcutHint.show()
+        self.settingShortcut = True
+        self.shortcutBtn.setProperty("class", "deleteBtn")
+        self.updateStyle(self.shortcutBtn)
+
+    def onRecordShortcutEnd(self, setOk=False):
+        self.remarkInput.setEnabled(True)
+        self.iconBtn.setEnabled(True)
+        self.okBtn.setEnabled(True)
+        self.cancelBtn.setEnabled(True)
+        if not setOk:
+            self.shortcutBtn.setText(_("Record"))
+            self.shortcut = []
+        self.shortcutHint.hide()
+        self.settingShortcut = False
+        self.shortcutBtn.setProperty("class", "")
+        self.updateStyle(self.shortcutBtn)
+
+    def keyPressEvent(self, event):
+        if not self.settingShortcut:
+            return
+        key = event.key()
+        name = QKeySequence(key).toString()
+        if key == Qt.Key_Control:
+            name = "Ctrl"
+        elif key == Qt.Key_Shift:
+            name = "Shift"
+        elif key == Qt.Key_Alt:
+            name = "Alt"
+        elif key == Qt.Key_Super_L:
+            name = "Super_L"
+        elif key == Qt.Key_Super_R:
+            name = "Super_R"
+        self.shortcut.append((key, name))
+        keys = "+".join([str(name) for v,name in self.shortcut])
+        self.shortcutBtn.setText(keys)
+
+    def keyReleaseEvent(self,event):
+        if not self.settingShortcut:
+            return
+        self.onRecordShortcutEnd(setOk = True)
+
+    def updateStyle(self, widget):
+        self.style().unpolish(widget)
+        self.style().polish(widget)
+        self.update()
 
-        
 
 class Plugin(Plugin_Base):
     '''
@@ -128,34 +201,29 @@ class Plugin(Plugin_Base):
             "currCode": "default",
             "customSendItems": [
                 {
+                    "text": "hello",
+                    "remark": "hello",
+                    "icon": "fa5.hand-paper"
+                },{
                     "text": "\\x01\\x03\\x03\\x03\\x03\\x01",
                     "remark": "pre",
-                    "icon": "ei.arrow-left"
-                },
-                {
+                    "icon": "ei.arrow-left",
+                    "shortcut": [[16777234, "Left"]]
+                },{
                     "text": "\\x01\\x04\\x04\\x04\\x04\\x01",
                     "remark": "next",
-                    "icon": "ei.arrow-right"
-                },
-                {
+                    "icon": "ei.arrow-right",
+                    "shortcut": [[16777236, "Right"]]
+                },{
                     "text": "\\x01\\x01\\x01\\x01\\x01\\x01",
                     "remark": "ok",
-                    "icon": "ei.ok"
-                },
-                {
+                    "icon": "fa.circle-o",
+                    "shortcut": [[16777220, "Return"]]
+                },{
                     "text": "\\x01\\x02\\x02\\x02\\x02\\x01",
                     "remark": "ret",
-                    "icon": "ei.return-key"
-                },
-                {
-                    "text": "",
-                    "remark": None,
-                    "icon": "fa.send"
-                },
-                {
-                    "text": "",
-                    "remark": None,
-                    "icon": "fa.send"
+                    "icon": "ei.return-key",
+                    "shortcut": [[16777216, "Esc"]]
                 }
             ]
         }
@@ -164,9 +232,36 @@ class Plugin(Plugin_Base):
             if not k in self.config:
                 self.config[k] = default[k]
         self.editingDefaults = False
-        self.codeGlobals = {"unpack": unpack, "pack": pack, "crc": crc}
+        self.codeGlobals = {"unpack": unpack, "pack": pack, "crc": crc, "encoding": self.configGlobal["encoding"], "print": self.print}
         self.encodeMethod = lambda x:x
         self.decodeMethod = lambda x:x
+        self.pressedKeys = []
+        self.keyModeClickTime = 0
+
+    def print(self, *args, **kw_args):
+        end = "\n"
+        start = "[MSG]: "
+        if "end" in kw_args:
+            end = kw_args["end"]
+        if "start" in kw_args:
+            start = kw_args["start"]
+        string = start  + " ".join(map(str, args)) + end
+        self.showReceiveDataSignal.emit(string)
+
+
+    class ModeButton(QPushButton):
+        onFocusIn = pyqtSignal(QFocusEvent)
+        onFocusOut = pyqtSignal(QFocusEvent)
+
+        def __init__(self, text, eventFilter, parent = None) -> None:
+            super().__init__(text, parent)
+            self.installEventFilter(eventFilter)
+
+        def focusInEvent(self, event):
+            self.onFocusIn.emit(event)
+
+        def focusOutEvent(self, event):
+            self.onFocusOut.emit(event)
 
     def onWidgetMain(self, parent, rootWindow):
         self.mainWidget = QSplitter(Qt.Vertical)
@@ -175,6 +270,13 @@ class Plugin(Plugin_Base):
         self.receiveWidget.setFont(font)
         self.receiveWidget.setLineWrapMode(TextEdit.NoWrap)
         self.clearBtn = QPushButton("")
+        self.keyBtneventFilter = self.ModeButtonEventFilter(self.onModeBtnKeyPressEvent, self.onModeBtnKeyReleaseEvent)
+        self.keyModeBtn = self.ModeButton(_("Key mode"), self.keyBtneventFilter)
+        layoutClearMode = QHBoxLayout()
+        layoutClearMode.addWidget(self.clearBtn)
+        layoutClearMode.addWidget(self.keyModeBtn)
+        clearModeWidget = QWidget()
+        clearModeWidget.setLayout(layoutClearMode)
         utils_ui.setButtonIcon(self.clearBtn, "mdi6.broom")
         self.addButton = QPushButton("")
         utils_ui.setButtonIcon(self.addButton, "fa.plus")
@@ -195,17 +297,17 @@ class Plugin(Plugin_Base):
         customSendItemsLayoutWrapper.setContentsMargins(0,0,0,0)
         cutomSendItemsWraper.setLayout(customSendItemsLayoutWrapper)
         # items container
-        customItems = QWidget()
+        self.customItems = QWidget()
         self.customSendItemsLayout = QVBoxLayout()
         self.customSendItemsLayout.setContentsMargins(0,0,0,0)
-        customItems.setLayout(self.customSendItemsLayout)
+        self.customItems.setLayout(self.customSendItemsLayout)
 
-        customSendItemsLayoutWrapper.addWidget(customItems)
+        customSendItemsLayoutWrapper.addWidget(self.customItems)
         customSendItemsLayoutWrapper.addWidget(self.addButton)
         customSendItemsLayoutWrapper.addStretch(0)
 
         self.mainWidget.addWidget(self.receiveWidget)
-        self.mainWidget.addWidget(self.clearBtn)
+        self.mainWidget.addWidget(clearModeWidget)
         self.mainWidget.addWidget(cutomSendItemsWraper0)
         self.mainWidget.setStretchFactor(0, 2)
         self.mainWidget.setStretchFactor(1, 1)
@@ -215,6 +317,38 @@ class Plugin(Plugin_Base):
         def clearReceived():
             self.receiveWidget.clear();self.clearCountSignal.emit()
         self.clearBtn.clicked.connect(clearReceived)
+        def keyModeOn(event):
+            self.keyModeBtn.setProperty("class", "deleteBtn")
+            utils_ui.updateStyle(self.mainWidget, self.keyModeBtn)
+            self.keyModeClickTime = time.time()
+            # show all shortcut
+            widgets = self.customItems.findChildren(QPushButton, "editRemark")
+            for i, w in enumerate(widgets):
+                shortcut = "+".join((name for v, name in self.config["customSendItems"][i]["shortcut"]))
+                w.setText(shortcut)
+                utils_ui.updateStyle(self.mainWidget, w)
+
+        def keyModeOff(event):
+            self.keyModeBtn.setProperty("class", "")
+            utils_ui.updateStyle(self.mainWidget, self.keyModeBtn)
+            self.keyModeClickTime = 0
+            # remove all preesed keys even them not release actually, to avoid window swith by ALT+TAB bug
+            self.pressedKeys = []
+            # hide all shortcut
+            widgets = self.customItems.findChildren(QPushButton, "editRemark")
+            for w in widgets:
+                w.setText("")
+
+        def keyModeTuggle():
+            if self.keyModeBtn.property("class") == "deleteBtn":
+                if time.time() - self.keyModeClickTime < 0.2:
+                    return
+                else:
+                    self.keyModeBtn.clearFocus()
+
+        self.keyModeBtn.onFocusIn.connect(keyModeOn)
+        self.keyModeBtn.onFocusOut.connect(keyModeOff)
+        self.keyModeBtn.clicked.connect(keyModeTuggle)
         return self.mainWidget
 
     def onWidgetSettings(self, parent):
@@ -231,6 +365,7 @@ class Plugin(Plugin_Base):
         self.codeItems.setEditable(True)
         self.codeWidget = PlainTextEdit()
         self.saveCodeBtn = QPushButton(_("Save"))
+        self.saveCodeBtn.setEnabled(False)
         self.deleteCodeBtn = QPushButton(_("Delete"))
         btnLayout = QHBoxLayout()
         btnLayout.addWidget(self.saveCodeBtn)
@@ -267,6 +402,7 @@ class Plugin(Plugin_Base):
         self.sendSettingsEscape.clicked.connect(lambda : self.bindVar(self.sendSettingsEscape, self.config, "sendEscape", bool))
         self.saveCodeBtn.clicked.connect(self.saveCode)
         self.deleteCodeBtn.clicked.connect(self.deleteCode)
+        self.codeWidget.onSave = self.saveCode
         return root
 
     def onWidgetFunctional(self, parent):
@@ -301,6 +437,11 @@ class Plugin(Plugin_Base):
                     config = json.load( f)
                     self.oldConfig = self.config.copy()
                     self.config.clear()
+                    if "plugins" in config and self.id in config["plugins"]: # global config file
+                        config = config["plugins"][self.id]
+                    if (not "plugin_id" in config) or not config["plugin_id"] == self.id:
+                        self.hintSignal.emit("error", _("Error"), _("config file format error"))
+                        return
                     for k in config:
                         self.config[k] = config[k]
                     def onClose(ok):
@@ -342,6 +483,45 @@ class Plugin(Plugin_Base):
         self.codeItems.setCurrentIndex(idx)
         self.selectCode(name)
         self.codeItems.currentIndexChanged.connect(self.onCodeItemChanged) # add here to avoid self.selectCode trigger
+        self.codeWidget.textChanged.connect(self.onCodeChanged)
+
+    class ModeButtonEventFilter(QObject):
+        def __init__(self, keyPressCb, keyReleaseCb) -> None:
+            super().__init__()
+            self.keyPressCb = keyPressCb
+            self.keyReleaseCb = keyReleaseCb
+
+        def eventFilter(self, obj, evt):
+            if evt.type() == QEvent.KeyPress:
+                # prevent default key events
+                self.keyPressCb(evt)
+                return True
+            elif evt.type() == QEvent.KeyRelease:
+                self.keyReleaseCb(evt)
+                return True
+            return False
+
+    def onModeBtnKeyPressEvent(self, event):
+        # send by shortcut
+        key = event.key()
+        self.pressedKeys.append(key)
+        for item in self.config["customSendItems"]:
+            if not "shortcut" in item:
+                continue
+            shortcut = item["shortcut"]
+            if len(shortcut) == len(self.pressedKeys):
+                same = True
+                for i in range(len(shortcut)):
+                    if shortcut[i][0] != self.pressedKeys[i]:
+                        same = False
+                        break
+                if same:
+                    self.sendCustomItem(item)
+
+    def onModeBtnKeyReleaseEvent(self, event):
+        key = event.key()
+        if key in self.pressedKeys:
+            self.pressedKeys.remove(key)
 
     def onKeyPressEvent(self, event):
         pass
@@ -349,7 +529,7 @@ class Plugin(Plugin_Base):
     def onKeyReleaseEvent(self, event):
         pass
 
-    def insertSendItem(self, item = {"text": "", "remark": None, "icon": None}, load = False):
+    def insertSendItem(self, item = None, load = False):
         # itemsNum = self.customSendItemsLayout.count() + 1
         # height = parameters.customSendItemHeight * (itemsNum + 1) + 20
         # topHeight = self.receiveWidget.height() + 100
@@ -358,6 +538,8 @@ class Plugin(Plugin_Base):
         # if height < 0:
         #     height = self.funcParent.height() // 3
         # self.customSendScroll.setMinimumHeight(height)
+        if not item:
+            item = {"text": "", "remark": None, "icon": None}
         if type(item) == str:
             item = {
                 "text": item,
@@ -376,8 +558,11 @@ class Plugin(Plugin_Base):
             send = QPushButton("")
         if (not "icon" in item) or not item["icon"]:
             item["icon"] = "fa.send"
+        if not "shortcut" in item:
+            item["shortcut"] = []
         utils_ui.setButtonIcon(send, item["icon"])
         editRemark = QPushButton("")
+        editRemark.setObjectName("editRemark")
         utils_ui.setButtonIcon(editRemark, "ei.pencil")
         editRemark.setProperty("class", "remark")
         cmd.setToolTip(text)
@@ -398,7 +583,10 @@ class Plugin(Plugin_Base):
         def changeRemark(idx, obj):
             if not "icon" in self.config["customSendItems"][idx]:
                 self.config["customSendItems"][idx]["icon"] = None
-            ok, remark, icon = EditRemarDialog(obj.text(), self.config["customSendItems"][idx]["icon"]).exec()
+            shortcut = []
+            if "shortcut" in self.config["customSendItems"][idx]:
+                shortcut = self.config["customSendItems"][idx]["shortcut"]
+            ok, remark, icon, shortcut = EditRemarDialog(obj.text(), self.config["customSendItems"][idx]["icon"], shortcut).exec()
             if ok:
                 obj.setText(remark)
                 if icon:
@@ -407,6 +595,7 @@ class Plugin(Plugin_Base):
                     obj.setIcon(QIcon())
                 self.config["customSendItems"][idx]["remark"] = remark
                 self.config["customSendItems"][idx]["icon"] = icon
+                self.config["customSendItems"][idx]["shortcut"] = shortcut
         editRemark.clicked.connect(lambda: changeRemark(self.customSendItemsLayout.indexOf(itemWidget), send))
         self.customSendItemsLayout.addWidget(itemWidget)
         if not load:
@@ -468,11 +657,10 @@ class Plugin(Plugin_Base):
         text = edit.text()
         edit.setToolTip(text)
         send.setToolTip(text)
-        item = {
+        self.config["customSendItems"][idx].update({
             "text": text,
             "remark": send.text()
-        }
-        self.config["customSendItems"][idx] = item
+        })
 
     def onCodeItemChanged(self):
         if self.editingDefaults:
@@ -487,7 +675,6 @@ class Plugin(Plugin_Base):
                 idx = self.codeItems.findText(name)
                 if idx >= 0:
                     self.codeItems.removeItem(idx)
-                print(self.codeItems.count(), self.codeItems.count() - 2)
                 self.codeItems.insertItem(self.codeItems.count() - 2, name)
                 self.config["code"][name] = defaultProtocols[name]
             self.codeItems.setCurrentIndex(0)
@@ -509,6 +696,8 @@ class Plugin(Plugin_Base):
         if ok:
             self.encodeMethod = e
             self.decodeMethod = d
+        self.saveCodeBtn.setText(_("Save"))
+        self.saveCodeBtn.setEnabled(False)
 
     def getEnDecodeMethod(self, code):
         func = lambda x:x
@@ -522,6 +711,21 @@ class Plugin(Plugin_Base):
             self.hintSignal.emit("error", _("Error"), msg)
         return False, func, func
 
+    def onCodeChanged(self):
+        changed = True
+        name = self.codeItems.currentText()
+        if name in self.config["code"]:
+            codeSaved = self.config["code"][name]
+            code = self.codeWidget.toPlainText()
+            if code == codeSaved:
+                changed = False
+        if changed:
+            self.saveCodeBtn.setText(_("Save") + " *")
+            self.saveCodeBtn.setEnabled(True)
+        else:
+            self.saveCodeBtn.setText(_("Save"))
+            self.saveCodeBtn.setEnabled(False)
+
     def saveCode(self):
         self.editingDefaults = True
         name = self.codeItems.currentText()
@@ -539,14 +743,16 @@ class Plugin(Plugin_Base):
             self.encodeMethod = e
             self.decodeMethod = d
             self.config["code"][name] = code
-        
+            self.saveCodeBtn.setText(_("Save"))
+            self.saveCodeBtn.setEnabled(False)
+
     def deleteCode(self):
         self.editingDefaults = True
         name = self.codeItems.currentText()
         itemsConfig = [self.codeItemCustomStr, self.codeItemLoadDefaultsStr]
         # QMessageBox.infomation()
         if name in itemsConfig or not name:
-            self.hintSignal.emit("warning", _("Warning"), _("Please select a code profile name first to delte"))
+            self.hintSignal.emit("warning", _("Warning"), _("Please select a code profile name first to delete"))
             self.editingDefaults = False
             return
         idx = self.codeItems.findText(name)
